/* Bitmap Ripper 1.0 - a simple program to extract pictures from binary data files,
if they are stored in 8, 16, 24 or 32 bits without compression.  You can interactively
adjust the image width until the picture looks right.  It also allows you to scan
through a file fairly quickly, looking for patterns in the raw binary - with experience
you can visually recognise sounds, images, database tables and other large binary
things.

Since this is merely crummy unsupported code written quickly,
it is released to the public domain.

1.0 Quick BeOS hack job by AGMS, November 2000.
*/

#include <Application.h>
#include <Window.h>
#include <View.h>
#include <Bitmap.h>
#include <FilePanel.h>
#include <Path.h>
#include <Alert.h>

#include <stdio.h>
#include <string.h>
#include <errno.h>

static FILE *DataFile;
static int BitmapWidthGuess = 1;
static int BytesPerPixel = 2;
static long FileOffset;
static bool FlipUpsideDown = true;


class BitmapRipperApp : public BApplication
{
public:
  BitmapRipperApp () : BApplication ("application/x-vnd.agmsmith.BitmapRipper") {};
  virtual void ReadyToRun();
};



class BitmapRipperWindow : public BWindow
{
public:
  BitmapRipperWindow ()
  : BWindow (BRect (30, 30, 670, 400), "BitmapRipper",
    B_TITLED_WINDOW, 0),
    m_BitmapPntr (NULL)
  {
  };

  ~BitmapRipperWindow ()
  {
    if (m_BitmapPntr != NULL)
      delete m_BitmapPntr;
    m_BitmapPntr = NULL;
  };

  bool QuitRequested ()
  {
    be_app->PostMessage (B_QUIT_REQUESTED);
    return true;
  };

  BBitmap *m_BitmapPntr;
};



class BitmapRipperView : public BView
{
public:
  BitmapRipperView (BRect FrameSize)
  : BView (FrameSize, "BitmapRipperView",
      B_FOLLOW_ALL, B_WILL_DRAW | B_FRAME_EVENTS),
    m_FilePanelPntr (NULL)
  {
    SetViewColor (B_TRANSPARENT_COLOR);
  };

  ~BitmapRipperView ()
  {
    if (m_FilePanelPntr != NULL)
    {
      delete m_FilePanelPntr;
      m_FilePanelPntr = NULL;
    }
  };

  void Draw (BRect updateRect);
  void FrameResized (float Width, float Height);
  void KeyDown (const char *Bytes, int32 NumBytes);
  void MouseDown (BPoint Point);
  void MessageReceived (BMessage* MessagePntr);
  void FillBitmapWithData (BBitmap *BitmapPntr);
  
  BFilePanel *m_FilePanelPntr; /* Gets file name to open from user. */
};



int main (int argc, char** argv)
{
  BitmapRipperApp theApp;

  theApp.Run();

  if (DataFile != NULL)
    fclose (DataFile);
  return 0;
}



void BitmapRipperApp::ReadyToRun ()
{
  BitmapRipperView   *MyView;
  BitmapRipperWindow *MyWin;

  MyWin = new BitmapRipperWindow;
  MyView = new BitmapRipperView (MyWin->Bounds ());

  MyWin->AddChild (MyView);
  MyWin->Show ();

  MyWin->Lock ();
  MyView->MakeFocus ();  /* We want this view to receive keystrokes. */
  MyWin->Unlock ();
}



void BitmapRipperView::KeyDown (const char *Bytes, int32 NumBytes)
{
  BitmapRipperWindow *WindowPntr;

  if (NumBytes == 1)
  {
    WindowPntr = (BitmapRipperWindow *) Window ();

    switch (Bytes[0])
    {
      case '.':
        FileOffset++; // Odd/even byte offset togglish.
        Invalidate ();
        break;

      case '1':
        BytesPerPixel = 1;
        Invalidate ();
        break;

      case '2':
        BytesPerPixel = 2;
        Invalidate ();
        break;

      case '3':
        BytesPerPixel = 3;
        Invalidate ();
        break;

      case '4':
        BytesPerPixel = 4;
        Invalidate ();
        break;
        
      case 'v':
        FlipUpsideDown = !FlipUpsideDown;
        Invalidate ();
        break;
        
      case B_UP_ARROW:
        FileOffset += BitmapWidthGuess * BytesPerPixel;
        Invalidate ();
        break;

      case B_DOWN_ARROW:
        FileOffset -= BitmapWidthGuess * BytesPerPixel;
        if (FileOffset < 0) FileOffset = 0;
        Invalidate ();
        break;

      case B_PAGE_UP:
        FileOffset += BitmapWidthGuess * 200 * BytesPerPixel;
        Invalidate ();
        break;

      case B_PAGE_DOWN:
        FileOffset -= BitmapWidthGuess * 200 * BytesPerPixel;
        if (FileOffset < 0) FileOffset = 0;
        Invalidate ();
        break;

      case B_RIGHT_ARROW:
        FileOffset -= BytesPerPixel;
        if (FileOffset < 0) FileOffset = 0;
        Invalidate ();
        break;

      case B_LEFT_ARROW:
        FileOffset += BytesPerPixel;
        Invalidate ();
        break;

      case B_HOME:
        if (--BitmapWidthGuess < 1) BitmapWidthGuess = 1;
        Invalidate ();
        break;

      case B_END:
        if (++BitmapWidthGuess > 6000) BitmapWidthGuess = 6000;
        Invalidate ();
        break;

      case B_INSERT:
        if (m_FilePanelPntr == NULL)
        {
          BMessenger Messenger (this); /* Send messages to me. */

          m_FilePanelPntr = new BFilePanel (
            B_OPEN_PANEL /* mode */,
            &Messenger /* target for event messages */,
            NULL /* starting directory */,
            B_FILE_NODE, /* what to open */
            false /* true for multiple selections */,
            NULL /* canned message */,
            NULL /* ref filter */,
            true /* true for modal */,
            true /* true to hide when done */);
        }
        if (m_FilePanelPntr != NULL)
          m_FilePanelPntr->Show (); /* Answer returned in B_REFS_RECEIVED. */
        break;
        
      case B_DELETE:
        if (WindowPntr->m_BitmapPntr != NULL)
          delete WindowPntr->m_BitmapPntr;
        WindowPntr->m_BitmapPntr = NULL;
        if (DataFile != NULL)
        {
          fclose (DataFile);
          DataFile = NULL;
        }
        Invalidate ();
        break;

      default:
        BView::KeyDown (Bytes, NumBytes);
    }
  }
  else
    BView::KeyDown (Bytes, NumBytes);
}


void BitmapRipperView::FillBitmapWithData (BBitmap *BitmapPntr)
{
  BRect   BitBounds;
  int     BytesPerRow;
  uint8  *FinalRowPntr;
  uint8  *FirstRowPntr;
  int     Height;
  uint32  InPixel;
  int     Length;
  int     MinWidth;
  uint32  OutPixel;
  uint8  *RowPntr;
  uint8   ScanLine [6000 * 4];
  uint8  *ScanLinePntr;
  uint32 *Uint32Pntr;
  int     Width;
  int     x;
  int     y;

  BitBounds = BitmapPntr->Bounds ();
  FirstRowPntr = (uint8 *) BitmapPntr->Bits ();
  Length = BitmapPntr->BitsLength ();
  memset (FirstRowPntr, 127, Length); // Fill it with gray.
  BytesPerRow = BitmapPntr->BytesPerRow ();
  FinalRowPntr = FirstRowPntr + Length - BytesPerRow;
  Height = 1 + (int) (BitBounds.bottom - BitBounds.top);
  Width = 1 + (int) (BitBounds.right - BitBounds.left);

  fseek (DataFile, FileOffset, SEEK_SET);

  MinWidth = Width;
  if (BitmapWidthGuess < MinWidth)
    MinWidth = BitmapWidthGuess;

  for (y = 0; y < Height; y++)
  {
    /* Set RowPntr to point to the start of the next row of pixels. */
    
    if (FlipUpsideDown)
    {
      RowPntr = FinalRowPntr;
      FinalRowPntr -= BytesPerRow; /* Reverse picture vertically. */
    }
    else
    {
      RowPntr = FirstRowPntr;
      FirstRowPntr += BytesPerRow;
    }

    if (fread (ScanLine, BitmapWidthGuess * BytesPerPixel, 1, DataFile) != 1)
      break; // Hit end of file.

    Uint32Pntr = (uint32 *) RowPntr;
    ScanLinePntr = ScanLine;

    for (x = 0; x < MinWidth; x++)
    {
      switch (BytesPerPixel)
      {
        default:
        case 1:
          /* Output pixel format is $AARRGGBB */
          /* Input pixel format is 8 bit gray. */

          InPixel = (ScanLinePntr[0]);
          ScanLinePntr += 1;

          OutPixel = InPixel;
          OutPixel |= (InPixel << 8);
          OutPixel |= (InPixel << 16);
          OutPixel |= (0xFF << 24);
          break;

        case 2:
          /* Output pixel format is $AARRGGBB */
          /* Input pixel format is %0rrr rrgg gggb bbbb */

          InPixel = (ScanLinePntr[0] | (ScanLinePntr [1] << 8));
          ScanLinePntr += 2;

          OutPixel = (InPixel & 0x1F) << 3;
          InPixel >>= 5;
          OutPixel |= ((InPixel & 0x1F) << 11);
          InPixel >>= 5;
          OutPixel |= ((InPixel & 0x1F) << 19);
          OutPixel |= (0xFF << 24);
          break;
          
        case 3:
          /* Output pixel format is $AARRGGBB */
          /* Input pixel format is $RRGGBB */

          InPixel = (ScanLinePntr[0] |
            (ScanLinePntr [1] << 8) |
            (ScanLinePntr [2] << 16));
          ScanLinePntr += 3;

          OutPixel = InPixel;
          OutPixel |= (0xFF << 24);
          break;
          
        case 4:
          /* Output pixel format is $AARRGGBB */
          /* Input pixel format is $AARRGGBB */

          InPixel = (ScanLinePntr[0] |
            (ScanLinePntr [1] << 8) |
            (ScanLinePntr [2] << 16) |
            (ScanLinePntr [2] << 24));
          ScanLinePntr += 4;

          OutPixel = InPixel;
          break;
          
        case 5:
          /* Output pixel format is $AARRGGBB */
          /* Input pixel format is %0rrr rrgg gggb bbbb */

          InPixel = (ScanLinePntr[0] | (ScanLinePntr [1] << 8));
          ScanLinePntr += 2;

          OutPixel = (InPixel & 0x1F) << 3;
          InPixel >>= 5;
          OutPixel |= ((InPixel & 0x1F) << 11);
          InPixel >>= 5;
          OutPixel |= ((InPixel & 0x1F) << 19);
          OutPixel |= (0xFF << 24);
          break;
          
      }
      *Uint32Pntr = OutPixel;
      Uint32Pntr++;
    }
  }
}


void BitmapRipperView::MouseDown (BPoint Point)
{
  MakeFocus (); /* We want to receive keystrokes. */
}


void BitmapRipperView::MessageReceived (BMessage* MessagePntr)
{
  entry_ref EntryRef;

  switch (MessagePntr->what)
  {
    case B_REFS_RECEIVED: /* Files from the file requestor have arrived. */
      if (MessagePntr->FindRef ("refs", &EntryRef) == B_OK)
      {
        BEntry MyEntry (&EntryRef, true /* traverse symlinks */);

        if (MyEntry.InitCheck () == B_OK)
        {
          BPath MyPath;

          if (MyEntry.GetPath (&MyPath) == B_OK)
          {
            if (DataFile != NULL)
              fclose (DataFile);

            /* All that to open the file. */
            
            DataFile = fopen (MyPath.Path (), "rb");

            if (DataFile == NULL)
            {
              BAlert *AlertPntr;
              int  ErrorCode = errno;
              char ErrorBuffer [B_PATH_NAME_LENGTH + 80];

              sprintf (ErrorBuffer, "Unable to open %s, "
                "error code $%X/%d (%s).",
                MyPath.Path (), ErrorCode, ErrorCode, strerror (ErrorCode));

              AlertPntr = new BAlert ("Error Message", ErrorBuffer,
                "OK", NULL, NULL, B_WIDTH_AS_USUAL, B_STOP_ALERT);
              if (AlertPntr != NULL)
                AlertPntr->Go ();
            }
            else /* File was successfully opened. */
            {
              BitmapRipperWindow *WindowPntr = (BitmapRipperWindow *) Window ();
    
              if (WindowPntr->m_BitmapPntr == NULL)
                WindowPntr->m_BitmapPntr = new BBitmap (Bounds (), B_RGB32);
              FileOffset = 0;
              BitmapWidthGuess = 640;
              rewind (DataFile);
            }
            Invalidate ();
          }
        }
      }
      break;

    default: /* Pass the buck. */
      BView::MessageReceived (MessagePntr);
      break;
  }
}


void BitmapRipperView::FrameResized (float Width, float Height)
{
  BitmapRipperWindow *WindowPntr;

  WindowPntr = (BitmapRipperWindow *) Window ();

  if (WindowPntr->m_BitmapPntr != NULL)
  {
    delete WindowPntr->m_BitmapPntr;
    WindowPntr->m_BitmapPntr = new BBitmap (BRect (0, 0, Width, Height), B_RGB32);
    Invalidate ();
  }
}


void BitmapRipperView::Draw (BRect updateRect)
{
  BBitmap            *BitmapPntr;
  BitmapRipperWindow *WindowPntr;

  WindowPntr = (BitmapRipperWindow *) Window ();
  BitmapPntr = WindowPntr->m_BitmapPntr;

  if (BitmapPntr != NULL)
  {
    FillBitmapWithData (BitmapPntr);
    MovePenTo (0, 0);
    DrawBitmap (BitmapPntr);
  }
  else
  {
    FillRect (Bounds (), B_SOLID_LOW);
    MovePenTo (10, 30);
    DrawString ("No bitmap available.  Use Insert to start.");
    MovePenTo (10, 50);
    DrawString ("Insert - open a new file.");
    MovePenTo (10, 70);
    DrawString ("Delete - close file, see instructions.");
    MovePenTo (10, 90);
    DrawString ("Left - scroll left 1 pixel.");
    MovePenTo (10, 110);
    DrawString ("Right - scroll right 1 pixel.");
    MovePenTo (10, 130);
    DrawString ("Up - scroll up 1 line.");
    MovePenTo (10, 150);
    DrawString ("Down - scroll down 1 line.");
    MovePenTo (10, 170);
    DrawString ("PageUp - scroll up several lines.");
    MovePenTo (10, 190);
    DrawString ("PageDown - scroll down several lines.");
    MovePenTo (10, 210);
    DrawString ("Home - Reduce bitmap width guess.");
    MovePenTo (10, 230);
    DrawString ("End - Increase bitmap width guess.");
    MovePenTo (10, 250);
    DrawString ("v - Flip vertically.");
    MovePenTo (10, 270);
    DrawString (". - Increase byte offset by 1, for odd/even fixups etc.");
    MovePenTo (10, 290);
    DrawString ("PrintScrn - Dump the screen to your Home directory.");
    MovePenTo (10, 310);
    DrawString ("1 2 3 4 - Set bytes per pixel to 1 (8 bit), 2 (16 bit), 3 (24 bit) or 4 (32 bit).");
    MovePenTo (10, 330);
    DrawString ("Bitmap Ripper version 1.0 by AGMS, November 2000.");
  }
}
